#include <iostream>
#include <iomanip>
#include <sstream>
#include <ctime>
#include <cmath>
#include <cstdlib>

//#define DEBUG 1
#include "Alibava.h"
#include "utils.h"
#include "GTimer.h"
#include "BitSet.h"
#include "path_utils.h"

class MutexPriv
{
    private:
        GMutex *mtx;
    public:
        MutexPriv(GMutex *m)
           : mtx(m)
        {
            if (mtx)
                g_mutex_lock(mtx);
        }
        ~MutexPriv()
        {
            if (mtx)
                g_mutex_unlock(mtx);
        }
};

const int Alibava::mxchan = 512;


std::string Alibava::find_dev_file()
{
    std::string out;
    std::vector<std::string> vout;
#ifdef __APPLE__
    PathUtils::find_file("cu.usbserial-", vout, "/dev");
#else
    PathUtils::find_file("alibava", vout, "/dev");
    #ifdef __linux__
    if (vout.empty())
        PathUtils::find_file("ttyUSB", vout, "/dev");
    #endif
#endif
    if (!vout.empty())
    {
        out = vout[0];
    }
    return out;
}

template<class T>
T reverse_bit(T v)
{
   T r = v; // r will be reversed bits of v; first get LSB of v
   int s = sizeof(v) * CHAR_BIT - 1; // extra shift needed at end

   for (v >>= 1; v; v >>= 1)
   {
      r <<= 1;
      r |= v & 1;
      s--;
   }
   r <<= s; // shift when v's highest bits are zero
   return r;
}

static std::string error_to_string(int code)
{
    std::string out;
    switch (code)
    {
        case 2:
            out = "SlowControlError";
            break;
        case 3:
            out = "ADCControlError1";
            break;
        case 4:
            out = "ADCControlError2";
            break;
        case 5:
            out = "ADCControlError";
            break;
        case 6:
            out = "LaserAcqError";
            break;
        case 8:
            out = "SDRAMError";
            break;
        default:
            out = "Unknown error code";
            break;
    }
    return out;
}

Alibava::Alibava(USBport &p) :
    do_stop(false), _error(0), _sample_size(10), _delay(3600),
    _chip_mask(3), _nchip(2), _debug(false),
    _fw_version(-1), _forced_fw(false), _fw_config(0),
    port(p)
{
    mtx = g_mutex_new();
}

Alibava::~Alibava()
{
    g_mutex_free(mtx);
}

bool Alibava::try_lock() const
{
	return g_mutex_trylock(mtx);
}

void Alibava::unlock()
{
	g_mutex_unlock(mtx);
}
void Alibava::lock()
{
	g_mutex_lock(mtx);
}


int Alibava::reopen()
{
    MutexPriv M(mtx);

    port.close();
    std::string device = find_dev_file();

#ifdef G_OS_UNIX
    if (device.empty())
    {
        err_msg(_("No device found in the system"));
        exit(1);
    }
#endif
    int rc = port.open(device);
    return rc;
}

int Alibava::open(const char *fnam)
{
    ALIBAVA_TRACE_CLASS("");
    return port.open(fnam);
}

int Alibava::reset(bool force)
{
    int rc, iok, itry;
    bool read_back;
    unsigned char buff=Connect;
    ALIBAVA_TRACE_CLASS("");

    PDEBUG("flushing the port\n");
    port.flush();

    PDEBUG("sending connexion code 0x0f\n");
    for (itry=0, read_back=false; itry<10; itry++)
    {
        rc = send_control(Connect);
        if (rc<0)
        {
            err_msg(_("Send connexion code failed. rc=%s"), error_string(iok).c_str());
            return rc;
        }

        rc = port.poll(100);
        if (rc==0)
        {
            rc = read(&iok, sizeof(unsigned char));
            if (rc>0)
                read_back = true;
            break;
        }
    }
    if (!read_back)
    {
        err_msg(_("Send connexion code failed. rc=%s"), error_string(iok).c_str());
        return iok;
    }

    /**
     * TODO: what happens with firmware if iok==waiting ?
     *       I guess it will, in any case, return the firmware version
     */
    unsigned char firmware_version = reverse_bit((unsigned char)iok) & 0xf;
    iok &= 0xf;
    if (!_forced_fw)
    	_fw_version = firmware_version;

    PDEBUG("Firmware version: %d\n", _fw_version);

    if (iok==Waiting)
    {
        if ( firmware() > 0)
        {
            return software_reset();
        }
        // Already in waiting
        PDEBUG("...MB does not need a reset\n");
        return 0;
    }

    if (iok!=Reset)
    {
        err_msg(_("Did not receive the proper Reset code: %x"), (unsigned short)buff);
        return StatusMismatch;
    }

    PDEBUG("sending handshake code\n");
    iok = send_control(Reset);
    if (iok < 0)
    {
        err_msg(_("Failed sending Reset. rc=%s"), error_string(iok).c_str());
        return iok;
    }

    // Check we are in Waiting
    unsigned char reg;
    iok = get_state(reg, Waiting, 100);
    if (iok<0)
    {
        err_msg(_("State machine did not end up in Waiting mode after reset"));
        return iok;
    }

    PDEBUG("...OK\n");
    return NoProblem;
}

int Alibava::software_reset()
{
    ALIBAVA_TRACE_CLASS("");
    int i, iok;

    //TODO: port.flush();

//    if ( firmware() < 1)
//        return reset();

    for ( int j=0; j<15; j++ )
    {
        for (i=0;i<4;i++)
        {
            iok = send_control(SftReset);
            if (iok!=NoProblem)
            {
                err_msg(_("Failed sending the software reset. rc=%s"), error_string(iok).c_str());
                return iok;
            }
        }
    }
    return reset();
}


int Alibava::config_firmware(unsigned short val)
{
    MutexPriv M(mtx);

    int rc;
    if (firmware()<2)
        return NoProblem;

    // Send the command
    rc = send_command(AlibavaConf, AlibavaConf);
    if (rc<0)
    {
        err_msg(_("Failed sending AlibavaConfig. rc=%s"), error_string(rc).c_str());
        return rc;
    }

    rc = send_word(val);
    if (rc==NoProblem)
        _fw_config = val;

    rc = send_data_word(EndOfData);
    if (rc<0)
    {
        err_msg(_("Failed sending EndOfData in ConfigFW. rc=%s"), error_string(rc).c_str());
    }
    // Check we are in Waiting
    unsigned char reg;
    rc = get_state(reg, Waiting);
    if (rc<0)
    {
        err_msg(_("State machine did not end up in Waiting mode after firmware config"));
    }

    return rc;
}

int Alibava::trigger_counter(unsigned short op)
{
    unsigned char reg;
    unsigned short ntrig;
    int rc;
    if (firmware()<2)
        return -1;

    // Send the command
    rc = send_command(TriggerCounter, TriggerCounter);
    if (rc<0)
    {
        err_msg(_("Failed sending TriggerCounter. rc=%s"), error_string(rc).c_str());
        return -1;
    }

    rc = send_control(op);
    rc = port.read(&ntrig, sizeof(unsigned short));

    rc = get_state(reg, None);
    if (reg!=Waiting)
    {
        err_msg(_("...final state is not Waiting. rc=%s"),error_string(rc).c_str());
        //rc = get_state(reg, None);
        return -2;
    }
    return ntrig;
}
int Alibava::get_trigger_counter()
{
    return trigger_counter(1);
}
void Alibava::reset_trigger_counter()
{
    int rc = trigger_counter(0);
    return;
}


int Alibava::beetle_header()
{
    int rc;
    unsigned char reg = BeetleConfig;

    // Send the command
    rc = send_command(BeetleConfig, BeetleConfig);
    if (rc<0)
    {
        err_msg(_("Failed sending BeetleConfig. rc=%s"), error_string(rc).c_str());
        return rc;
    }
    if ( firmware()>0 )
    {
        rc = send_control(mask());
        if ( rc != NoProblem )
            return rc;
    }
    return rc;
}
int Alibava::beetle_config(int nval, short mask_value, unsigned short *values)
{
    MutexPriv M(mtx);

    ALIBAVA_TRACE_CLASS("");

    // Set the value of the mask
    if (firmware()>0)
        mask(mask_value);
    else
        _nchip = 2;

    int rc = beetle_header();
    if (rc<0 || rc != NoProblem)
        return rc;

    // Write the values
    bool failed = false;
    int i=0;
    PDEBUG("Start loop\n");
    for (i=0; i<nval; i++)
    {
        PDEBUG("...index %d\n", i);
        rc = send_data_word(values[i]);
        if (rc<0)
        {
            failed = true;
            err_msg(_("Failed sending Beetle config word %d. rc=%s"), i, error_string(rc).c_str());
            break;
        }
    }
    if (failed)
    {
        PDEBUG("...failed\n");
        return rc;
    }
    /**
     * Initialize threshold TpSelect
     * We set them all to 0.
     */
    if (  firmware()>1 )
    {
        rc = send_data_word(0x1600);
        if (rc<0)
        {
            failed = true;
            err_msg(_("Failed sending Beetle TpSelect register chan=%d. rc=%s"), i, error_string(rc).c_str());
            return rc;
        }
    }

    return beetle_trailer();
}

int Alibava::beetle_config_mask(int chipid, const BitSet &B)
{
    MutexPriv M(mtx);
    ALIBAVA_TRACE_CLASS("");

    if ( firmware()<2 )
        return NoProblem;

    int rc, old_mask = mask();
    int new_mask = 1<<chipid;

    mask(new_mask);
    rc = beetle_header();
    mask(old_mask);
    if (rc<0 || rc != NoProblem)
        return rc;

    for (int iw=0; iw<16;iw++)
    {
        rc = send_data_word( 0x1500 | B.Word(iw) );
        if (rc<0)
        {
            err_msg(_("Failed sending Beetle mask register chan=%d. rc=%s"), iw, error_string(rc).c_str());
            break;
        }
    }

    return beetle_trailer();
}

int Alibava::beetle_config_trim(int chipid, const unsigned char *val)
{
    MutexPriv M(mtx);
    ALIBAVA_TRACE_CLASS("");

    if ( firmware()<2 )
        return NoProblem;

    int rc, old_mask = mask();
    int new_mask = 1<<chipid;

    mask(new_mask);
    rc = beetle_header();
    mask(old_mask);
    if (rc<0 || rc != NoProblem)
        return rc;

    for (int ichan=0; ichan<128; ichan++)
    {
        rc = send_data_word(0x1400 | val[ichan]);
        if (rc<0)
        {
            err_msg(_("Failed sending Beetle trim value chan=%d. rc=%s"), ichan, error_string(rc).c_str());
            break;
        }
    }

    return beetle_trailer();
}

int Alibava::beetle_config_comparator(int chipid, unsigned char thrs, unsigned char delta, unsigned char vrc, unsigned char icomp)
{
    ALIBAVA_TRACE_CLASS("");
    if (!firmware())
        return NoProblem;

    int rc = 0;
    unsigned short values[] = {
                               0x0600 | (icomp & 0xff),
                               0x0700 | (delta & 0xff),
                               0x0800 | (thrs & 0xff),
                               0x0900 | (vrc & 0xff)
    };
    int nval = sizeof(values)/sizeof(unsigned short);
    int old_mask = mask();
    int new_mask = 1<<chipid;

    rc = beetle_config(nval, new_mask, values);

    mask(old_mask);

    return rc;
}

int Alibava::beetle_trailer()
{
    unsigned char reg;
    int rc;

    // Check that it is waiting for new data
    rc = get_state(reg, NewData);
    if (rc<0)
    {
        err_msg(_("Failed checking for NewData: rc=%s"),error_string(rc).c_str());
        return rc;
    }

    // send end of data
    if ( (rc = send_word(EndOfData))<0)
    {
        err_msg(_("Failed sending EndOfData: rc=%s"),error_string(rc).c_str());
        return rc;
    }

    rc = get_state(reg, Waiting);
    if (reg!=Waiting)
    {
        PDEBUG("...final state is not Waiting\n");
        err_msg(_("...final state is not Waiting. rc=%s"), error_string(rc).c_str());
        //rc = get_state(reg, None);
        return StatusMismatch;
    }

    PDEBUG("...OK\n");
    return NoProblem;
}


int Alibava::wait_for_data(unsigned int size)
{
    ALIBAVA_TRACE_CLASS("");
    unsigned int sz = 0;
    double poll_delay = 100.0;
    AlibavaTimer tim;

    tim.start();
    do {
        sz =  port.wait_for_data(_debug ? -1 : poll_delay);// port.data_available(size);
        if ( sz>=size )
            break;

        g_usleep(1);
    } while (tim() < 5.0 && sz<size/2.0 );


    if ( sz<size/2.0 )
    {
        err_msg(_("Timeout after %3.1f sec. Found %d [%d]"), tim(),sz,size);
        return TimeOut;
    }
    return NoProblem;
}

/**
 * read_data: Reads the data in calibration mode.
 */
int Alibava::read_event_data(StateCode code, EventData *data)
{
    ALIBAVA_TRACE_CLASS("");

    int rc;
    if (wait_for_data(data->size_calib())<0)
        return TimeOut;

    rc = data->read_data(port);
    PDEBUG("Reading event data: %d\n",rc);
    if (rc>0)
    {
        ReceiverList::iterator ip;
        for (ip=receivers.begin(); ip!=receivers.end(); ip++)
            (*ip)->receive(data);
    }
    return rc;
}

int Alibava::do_calibration(StateCode code, unsigned short charge, unsigned short delay, int nevts)
{
    ALIBAVA_TRACE_CLASS("");

    int rc;
    unsigned char reg = code;
    // Send the command
    rc = send_command(code, code);
    if (rc<0)
        return -1;

    // Loop as many times as events
    int ievt;
    bool failed = false;
    EventData *data = EventData::factory( nchip(), firmware(), mask());
    if (code==LaserSync)
        data->value(delay);
    else
    {
        data->value( delay << 16 | charge );
    }

    for (ievt=0; ievt<nevts && !do_stop; ievt++)
    {
        PDEBUG("Event %d/%d\n",ievt,nevts);

        // Check for new data
        rc = check_for_data();
        if (rc<0 && rc != StatusMismatch)
        {
            err_msg(_("Error checking for new data state: rc=%s evt=%d"), error_string(rc).c_str(), ievt);
            failed = true;
            break;
        }

        if (code == LaserSync )
        {
            // Write the register
            rc = send_word(delay);
            if (rc<0)
            {
                err_msg(_("Failed sending LaserSync delay %d with rc=%x"), delay, rc);
                return rc;
            }
        }
        else
        {
            if ( firmware() > 0 )
            {
                // Write the register
                rc = send_word(delay);
                if (rc<0)
                {
                    err_msg(_("Failed sending Calibration delay %d with rc=%x"), delay, rc);
                    return rc;
                }
            }
            // Write the register
            rc = send_word(charge);
            if (rc<0)
            {
                err_msg(_("Failed sending Calibration charge %d with rc=%x"), charge, rc);
                return rc;
            }
        }

        // Get the ADC data

        // wait for data available
        rc = port.poll(_debug ? -1 : 10000);
        if (rc)
        {
            failed = true;
            err_msg(_("Poll failed"));
            break;
        }

        // read the data buffer
        if ( (rc=read_event_data(code, data))<0)
        {
            failed=true;
            err_msg(_("read_event_data failed"));
            break;
        }
    }
    delete data;


    // check
    if (failed)
    {
        err_msg(_("...failed with rc=%x"), rc);
        return rc;
    }

    // Read the pending NewData and send an EndOfData
    if ( (rc = get_state(reg, NewData))<0)
    {
        std::ostringstream ostr;
        if (rc==BoardError)
        {
            err_msg(_("Board error: %s"), error_string(rc).c_str());
        }
        else
        {
            err_msg(_("Failed checking for new data. rc=%s"), error_string(rc).c_str());
        }
    }
    if ( (rc = send_word(EndOfData))<0)
    {
        err_msg(_("Failed sending EndOfData. rc=%s"), error_string(rc).c_str());
    }

    rc = get_state(reg, None);
    if (reg!=Waiting)
    {
        err_msg(_("...final state is not Waiting. rc=%s"), error_string(rc).c_str());
        // TODO: figure out why we did this
        //rc = get_state(reg, None);
        return StatusMismatch;
    }

    PDEBUG("...OK\n");
    return NoProblem;
}

int Alibava::do_acquisition(unsigned short nevts, StateCode code)
{
    ALIBAVA_TRACE_CLASS("");

    int rc=0, ntrigger=0;
    unsigned char st;

    if (nevts <= 0)
        return -1;

    if (nevts > 64776)
        nevts = 64776;

    _sample_size = nevts;
    // Send the command
    rc = send_command(code, code);
    if (rc<0)
    {
        err_msg(_("Failed sending command %x with rc=%s"), code, error_string(rc).c_str());
        return rc;
    }
    // Write the number of samples that we want
    if ( (rc=send_word(nevts))<0)
    {
        err_msg(_("Failed sending number of events with rc=%s"),error_string(rc).c_str());
        return rc;
    }

    // now wait for the end of it
    AlibavaTimer tim;
    tim.start();
    do
    {
        rc = get_state(st,None, 1000, false);
        if ( rc==TimeOut && tim()>delay() )
        {
            software_reset();
            break;
            //rc=NoProblem;
        }
        if (do_stop)
        {
            return rc;
        }

    } while (rc==TimeOut && tim()<nevts*delay() );



    if (rc==TimeOut)
    {
        err_msg(_("Time out while waiting for Alibava to acquire the data."));
        return rc;
    }
    if (st!=Waiting)
    {
        // We got an error
        int error_code = (rc & 0xf0)>4;
        if (error_code)
        {
            err_msg(_("Error code: %x [%s]"), rc, error_to_string(error_code).c_str());
            rc = BoardError;
        }
        else
        {
            err_msg(_("Got error code %x while expecting Waiting"), rc);
            rc = StatusMismatch;
        }
        // now we get the Waiting (hopefully)
        get_state(st);
        if (st!=Waiting)
        {
            err_msg(_("Not getting Waiting after acquisition. rc=%s"), error_string(rc).c_str());
            err_msg(_("...The error we get is rc=%x"), st);
        }
        else
            rc = NoProblem;
    }
    else
    {
        rc = NoProblem;
    }


    return rc;
}


unsigned short nevt_mask(unsigned short maxevt)
{
	unsigned short val = (unsigned short)( pow(2, int( 0.5 + log(double(maxevt))/log(2.0) ))-1);
	return val;
}

/**
 * read_data: Reads the data in Laser, Pedestal and RS mode
 */
int Alibava::read_data(StateCode code)
{
    ALIBAVA_TRACE_CLASS("");

    int rc=0;
    char c;
    unsigned short ievt, nevt, nbit;
    unsigned char st;

    // Send the command
    rc = send_command(code, code);
    if (rc<0)
    {
        err_msg(_("Failed sending command %x with rc=%s"), code, error_string(rc).c_str());
        return rc;
    }

    nevt = 0;
    rc = port.read(&nevt, sizeof(unsigned short));
    if (rc<=0)
    {
        err_msg(_("Failed reading nevt from the port."));
        // TODO: figure out how to flush the input buffer
        return rc;
    }


    rc = port.read(&nbit, sizeof(unsigned short));
    if (rc<=0)
    {
        err_msg(_("Failed reading the number of bits from the port."));
        // TODO: figure out how to flush the input buffer
        port.flush();
        return rc;
    }

    /// Send the acknowledge
    rc = send_control(code);

    if (nevt == 0)
    {
        // we should be in WAITING..
        rc = get_state(st,Waiting);
        return NoProblem;
    }
//    if ( nevt != _sample_size )
//    {
//        err_msg( _("inconsistent sample size: %d != %d"), _sample_size, nevt);
//        return ReadError;
//    }



    EventData *evt = EventData::factory( nchip(), firmware(), mask());
    for (ievt=0; ievt<nevt; ievt++)
    {
        /**
         * Check for data availability
         */

        if (wait_for_data(evt->size_data())<0)
        {
            //if ( ievt+1 != nevt )
            	err_msg(_("Timeout waiting for data"));

			port.flush();
            info_msg(_("... Events read from buffer: %d/%d [%d]\n"), ievt, nevt, port.data_available(0));
            delete evt;
            return ievt; // I used to return TimeOut here.

        }

        port.read(&evt->get_block()->time, sizeof(unsigned int));
        port.read(&evt->get_block()->temp, sizeof(unsigned short));
        rc = evt->read_data(port); //port.read(evt->ptr_data(), evt->size_data());
        if (rc<=0)
        {
            err_msg(_("Error reading data"));
            delete evt;
            return ReadError;
        }

        rc = 8*rc + 48;
        if (rc<nbit)
        {
            err_msg(_("Error reading data: size mismatch (ievt=%d) read %d expected %d"), ievt, rc, nbit);
            port.flush();
            return DataSizeMismatch;
        }
        else
        {
            ReceiverList::iterator ip;
            for (ip=receivers.begin(); ip!=receivers.end(); ip++)
                (*ip)->receive(evt);
        }
    }
    delete evt;
    //std::cout << "Quedan " << port.data_available(0) << " bytes por leer" << std::endl;
    // Now check for Waiting state
    get_state(st);
    if (st!=Waiting)
    {
        err_msg(_("Not getting Waiting after reading data."));
        return StatusMismatch;
    }

    return ievt;
}


int Alibava::trigger(TriggerType type, TriggerSetup setup, unsigned short thrsA,
                     unsigned short thrsB, unsigned short thrsC, unsigned short thrsD)
{
    MutexPriv M(mtx);
    ALIBAVA_TRACE_CLASS("");
    int rc;
    unsigned char reg = type<<1 | setup;

    PDEBUG("Configuring Trigger\n");
    // Send the command
    rc = send_command(TriggerInConfig, TriggerInConfig);
    if (rc<0)
    {
        err_msg(_("Failed sending TriggerInConfig with rc=%s"), error_string(rc).c_str());
        return rc;
    }

    // send the configuration
    PDEBUG("...trigger code: %x\n", reg);
    rc = send_control(reg);
    if (rc<0)
    {
        err_msg(_("Failed sending trigger conf. %x with rc=%s"), reg, error_string(rc).c_str());
        return rc;
    }

    // Now send the thresholds
    PDEBUG("...thrsA: %x\n", thrsA);
    if ( (rc=send_data_word(thrsA & 0xFFF)) < 0)
    {
        err_msg(_("Failed sending thrsA. rc=%s"),error_string(rc).c_str());
        return rc;
    }

    PDEBUG(_("...thrsB: %x\n"), thrsB);
    if ( (rc=send_data_word( 0x1000 | (thrsB & 0xFFF))) < 0)
    {
        err_msg(_("Failed sending thrsB. rc=%s"), error_string(rc).c_str());
        return rc;
    }

    PDEBUG(_("...thrsC: %x\n"), thrsC);
    if ( (rc=send_data_word( 0x2000 | (thrsC & 0xFFF))) < 0)
    {
        err_msg(_("Failed sending thrsC. rc=%s"), error_string(rc).c_str());
        return rc;
    }

    PDEBUG(_("...thrsD: %x\n"), thrsD);
    if ( (rc=send_data_word( 0x3000 | (thrsD & 0xFFF))) < 0)
    {
        err_msg(_("Failed sending thrsD. rc=%s"), error_string(rc).c_str());
        return rc;
    }

    // Read the pending NewData and send an EndOfData
    if ( (rc = get_state(reg, NewData)) < 0)
    {
        err_msg(_("Failed checking for NewData with rc=%s"), error_string(rc).c_str());
    }
    if ( (rc = send_word(EndOfData)))
    {
        err_msg(_("Failed sending EndOfData with rc=%s"),error_string(rc).c_str());
    }
    rc = get_state(reg, None);
    if (reg!=Waiting)
    {
        err_msg(_("...final state is not Waiting. rc=%s"),error_string(rc).c_str());
        //rc = get_state(reg, None);
        return StatusMismatch;
    }

    PDEBUG(_("...OK\n"));

    return NoProblem;
}

int Alibava::set_laser_delay(int delay)
{
    ALIBAVA_TRACE_CLASS("");
    /**
     * We have to make a 1 event laser scan to set the delay
     */
    return laser_sync(delay, 1);
}

int Alibava::laser_sync(unsigned short delay, int nevts) // ...........................LaserSync..............
{
    ALIBAVA_TRACE_CLASS("");
    MutexPriv M(mtx);
    return do_calibration(LaserSync, 0, delay, nevts);
}

int Alibava::calibration(unsigned int charge, unsigned int delay, int nevts)
{
    ALIBAVA_TRACE_CLASS("");
    MutexPriv M(mtx);
    return do_calibration(Calibration, charge/1024, delay, nevts);
}

int Alibava::laser_acq(unsigned short nevts) // ...........................Laser Acquisition..........
{
    ALIBAVA_TRACE_CLASS("");
    MutexPriv M(mtx);
    return do_acquisition(nevts, LaserAcq);
}

int Alibava::laser_reading() // ...........................Laser Reading...........
{
    ALIBAVA_TRACE_CLASS("");
    MutexPriv M(mtx);
    return read_data(LaserReading);
}

int Alibava::rad_source_acq(unsigned short nevts) // ...................Rad Sourse Acquisition.........
{
    ALIBAVA_TRACE_CLASS("");
    MutexPriv M(mtx);
    return do_acquisition(nevts, RadSourceAcq);
}

int Alibava::rad_source_reading() // .................Rad Sourse Reading...........
{
    ALIBAVA_TRACE_CLASS("");
    MutexPriv M(mtx);
    return read_data(RadSourceReading);
}

int Alibava::pedestal_acq(unsigned short nevts)
{
    ALIBAVA_TRACE_CLASS("");
    MutexPriv M(mtx);
    return do_acquisition(nevts, PedestalAcq);
}

int Alibava::pedestal_reading()
{
    ALIBAVA_TRACE_CLASS("");
    MutexPriv M(mtx);
    return read_data(PedestalReading);
}

int Alibava::get_state(unsigned char &val, StateCode check, int tmout, bool notify)
{
    ALIBAVA_TRACE_CLASS("");
    int iok;
    unsigned char reg;

    val = None;
    // Now get the response
    PDEBUG("Getting the state code\n");
    if ( tmout<=0 )
        iok=0;
    else
        iok = port.poll(_debug ? -1 : tmout);

    if (iok==0)
    {
        iok = read(&reg, sizeof(unsigned char));
        if (iok<=0)
        {
            if (notify)
            {
                err_msg(_("Getting Status code failed"));
                Verbose::backtrace();
            }
            return ReadError;
        }
    }
    else if (iok>0)
    {
        if (notify)
        {
            err_msg(_("Getting status code timed out "));
            Verbose::backtrace();
        }
        return TimeOut;
    }
    else
    {
        if (notify)
        {
            err_msg(_("Getting status code failed"));
            Verbose::backtrace();
        }
        return ReadError;
    }
    PDEBUG(_("...code=%x\n"),reg);
    val = reg;

    if (check)
    {
        if (reg!=check)
        {
            if (notify)
            {
                err_msg(_("Status mismatch %x != %x"), reg, check);
                Verbose::backtrace();
            }
            return StatusMismatch;
        }
    }

    return NoProblem;
}

int Alibava::send_control(unsigned char cmd)
{
    ALIBAVA_TRACE_CLASS("");
    int iok;
    unsigned char buff = cmd;

    PDEBUG(_("Sending control word %d\n"), cmd);
    iok = write(&buff, sizeof(unsigned char));
    if (iok<=0 || !port)
    {
        PDEBUG(_("...failed\n"));
        err_msg(_("Failed sending control word %x"),cmd);
        return WriteError;
    }
    return NoProblem;
}

int Alibava::send_command(StateCode cmd, StateCode check, int tmout)
{
    ALIBAVA_TRACE_CLASS("");
    int iok;
    unsigned char buff = cmd;

    PDEBUG(_("Sending command %d\n"), cmd);
    iok = send_control(cmd);
    if (iok<0)
        return iok;

    // Check that it is the same
    iok = get_state(buff, check, tmout);
    if (iok<0)
        return iok;
    else
        return buff;
}

int Alibava::send_word(const unsigned short &val)
{
    ALIBAVA_TRACE_CLASS("");
    int iok;

    PDEBUG(_("Sending data word %x\n"), val);
    iok = write((void *)&val, sizeof(unsigned short));
    if (iok<=0)
    {
        PDEBUG(_("...failed\n"));
        if (iok<0)
        {
            err_msg(_("Failed sending data word %x"), val);
        }
        return WriteError;
    }
    return NoProblem;
}

int Alibava::check_for_data()
{
	ALIBAVA_TRACE_CLASS("");
    int rc;
    unsigned char reg;
    _error = 0;
    rc = get_state(reg, NewData);
    if (rc<0 && rc != StatusMismatch)
    {
        PDEBUG(_("rc = %d\n"),rc);
        return rc;
    }

    if (rc == StatusMismatch)
    {
        int icode = (reg >> 4);
        if (icode>1)
        {
            err_msg(_("SlowControl error"));
            rc = BoardError;
            _error = reg;
        }
        else
        {
            //err_msg(_("Inconsistent state. Should be %x instead of %x"), NewData, reg);
            rc = StatusMismatch;
        }
        PDEBUG(_("rc = %d\n"),rc);
        return rc;
    }
    PDEBUG(_("rc = %d\n"),rc);
    return rc;
}

int Alibava::send_data_word(unsigned short val)
{
    int rc;

    // Check for data
    if ( (rc = check_for_data())<0)
        return rc;

    // send the word
    rc = send_word(val);

    return rc;
}

std::string Alibava::error_string(int rc) const
{
    std::string out;
    switch (rc)
    {
        case NoProblem:
            out="No problem";
            break;
        case ReadError:
            out="Read Error";
            break;
        case WriteError:
            out="Write Error";
            break;
        case TimeOut:
            out="Timeout";
            break;
        case StatusMismatch:
            out="Status mismatch";
            break;
        case BoardError:
            out = "Board Error";
            break;
        case DataSizeError:
            out = "Data Size Error";
            break;
        case DataSizeMismatch:
            out = "Data Size Mismatch";
            break;
        default:
            out=std::string(strerror(errno)) + "?????";
            break;
    }
    return out;
}

void Alibava::remove_data_receiver(DataReceiver *d)
{
    ReceiverList::iterator ip;
    for (ip=receivers.begin(); ip!=receivers.end(); ip++)
    {
    	if (*ip == d)
    	{
    		receivers.erase(ip);
    		break;
    	}
    }
}

void Alibava::define_scan(int npts, double from, double to)
{
    ReceiverList::iterator ip;
    for (ip=receivers.begin(); ip!=receivers.end(); ip++)
    {
        (*ip)->define_scan(npts, from, to);
    }
}
void Alibava::end_of_scan_point(double value, unsigned int ntrg)
{
    ReceiverList::iterator ip;
    for (ip=receivers.begin(); ip!=receivers.end(); ip++)
    {
        (*ip)->end_of_scan_point(value, ntrg);
    }
}

