#include <glib.h>
#ifdef HAVE_CONFIG_H
#ifdef G_OS_WIN32
#include "config_win.h"
#else
#include <config.h>
#endif
#else
#warning "NO config"
#endif

#ifdef HAVE_GIO

#include <iostream>
#include <map>
#include <cstdlib>
#include <cstdio>
#include <cstdarg>
#include <cerrno>
#include <csignal>
#include <cstring>
#include <ctime>
#include <cmath>

#include <unistd.h>
#include <glib.h>
#include <gio/gio.h>

#include "Fifo.h"
#include "USBemulator.h"
#include "utils.h"
#include "path_utils.h"
#include "alibava-server.h"

#ifdef G_OS_WIN32
#include <winsock.h>
#define HANDLE unsigned long
#else
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#endif
void __g_gpid_kill( GPid pid)
{
#ifdef G_OS_WIN32
    if (!TerminateProcess(reinterpret_cast<void *>( pid ), 0))
    {
        gchar *msg = g_win32_error_message(GetLastError());
        g_warning("Could not kill child process: %s", msg ? msg : "(null)");
        g_free(msg);
    }
#else /* !G_OS_WIN32 */
    if (kill(pid, SIGKILL))
    {
        g_warning("Could not kill child process: %s", g_strerror(errno));
    }
#endif /* G_OS_WIN32 */
}


/*
 * This function creates a connection to the server.
 */
static GSocket *connect_server(char type, const char *name)
{
    int opt, sockfd;
    gchar recv_buf[256];
    gsize sz = strlen(name)+1;
    GSocket *fd;
    GError *error = 0;

    /* Create first a socket that will identify us in the server. */
    fd = g_socket_new(G_SOCKET_FAMILY_IPV4,
                      G_SOCKET_TYPE_STREAM,
                      G_SOCKET_PROTOCOL_DEFAULT,
                      &error);
    if (fd == NULL)
    {
        g_printerr ("%s: %s\n", "create_socket: ", error->message);
        return 0;
    }

    opt = 1;
    sockfd = g_socket_get_fd(fd);
    setsockopt (sockfd, IPPROTO_TCP,
                    TCP_NODELAY, (char *) &opt, sizeof (opt));
    setsockopt (sockfd, SOL_SOCKET,
                    SO_REUSEADDR, (char *) &opt, sizeof (opt));




    GSocketConnectable *connectable = g_network_address_parse ("127.0.0.1",
                                                               SERVER_PORT,
                                                               &error);
    if (connectable == NULL)
    {
        g_printerr ("%s: %s\n", "connect_server", error->message);
        return 0;
    }
    GSocketAddressEnumerator *enumerator =  g_socket_connectable_enumerate (connectable);
    while (true)
    {
        GSocketAddress *address = g_socket_address_enumerator_next (enumerator, 0, &error);
        if (!address)
        {
            if (error)
            {
                g_printerr ("%s: %s\n", "connect_server", error->message);
            }
            return 0;
        }
        if (g_socket_connect (fd, address, 0, &error))
            break;

        error = NULL;

        g_object_unref (address);

    }
    g_object_unref (enumerator);
    g_object_unref (connectable);

    g_socket_set_blocking (fd, TRUE);
    return fd;
}


static void kill_daemon()
{
#ifdef G_OS_UNIX
    FILE *pf;
    char line[128];
#ifdef __APPLE__
    sprintf(line,"/bin/ps -c");
#else
    sprintf(line,"/bin/ps -C %s", DAEMON_NAME);
#endif
    pf = popen(line, "r");
    if (!pf)
        return;

    while (fgets(line, sizeof(line), pf))
    {
        if (strstr(line, DAEMON_NAME))
        {
            int pid;
            if ( sscanf(line,"%d",&pid)==1 )
            {
                kill(pid, 9);
            }
        }
    }
    pclose(pf);
#endif
}

static void
cb_child_watch( GPid  pid,
                gint  status,
                gpointer data )
{

    /* Close pid */
    g_spawn_close_pid( pid );
}

static gboolean
cb_out_watch( GIOChannel   *channel,
              GIOCondition  cond,
              gpointer      data )
{
    gchar *string;
    gsize  size;

    if( cond == G_IO_HUP )
    {
        g_io_channel_unref( channel );
        return( FALSE );
    }

    g_io_channel_read_line( channel, &string, &size, NULL, NULL );
    if ( size )
    {
        g_print("[Server] %s", string);
        g_free( string );
        return( TRUE );
    }
    else
        return FALSE;
}

static GPid check_for_daemon(int firmware)
{
    char line[512], CWD[512];
    char *cmmd;
    GError *error = 0;
    GPid pid;
    gint argc;
    gchar **argv;
    gchar *buff;
    gsize buff_len, tpos;
    gint out;
    GIOChannel *out_ch;
    extern gchar *__daemon_path__;

    kill_daemon();

    std::string daemon(__daemon_path__);
    if (!PathUtils::exists(daemon))
        daemon = PathUtils::find_bin( PathUtils::join( DAEMON_PATH, PathUtils::base_name(daemon)  ) );

    std::string cwd = PathUtils::getcwd();
    std::string path = PathUtils::dir_name(daemon);
    std::string prgname = PathUtils::base_name(daemon);
    chdir(path.c_str());
    cmmd = getenv("ALIBAVAD");
    if (!cmmd)
    {
        sprintf(line, "./%s --daemon=no --firmware-version=%d", prgname.c_str(), firmware);
        g_shell_parse_argv(line, &argc, &argv, &error);
        if ( error )
        {
            g_printerr("check_for_daemon: %s\n", error->message);
            g_error_free(error);
            exit(1);
        }
        cmmd = line;
    }
    else
    {
        g_shell_parse_argv(cmmd, &argc, &argv, &error);
        if ( error )
        {
            g_printerr("check_for_daemon (ALIBAVAD): %s\n", error->message);
            g_error_free(error);
            exit(1);
        }
    }
    /*
     * TODO: which is the workding directory of alibavad ?
     */
    if  (! g_spawn_async_with_pipes(NULL, argv, NULL,
                                    G_SPAWN_DO_NOT_REAP_CHILD,
                                    NULL, NULL,
                                    &pid, NULL, NULL,&out, &error) )
    {
        err_msg("Could not start daemon: %s\n[%s]\n", error->message, cmmd);
        g_error_free(error);
        exit(1);
    }
    chdir( cwd.c_str() );
    g_child_watch_add( pid, (GChildWatchFunc)cb_child_watch, NULL );

#ifdef G_OS_WIN32
    out_ch = g_io_channel_win32_new_fd( out );
#else
    out_ch = g_io_channel_unix_new( out );
#endif

    error = 0;
    buff_len = tpos = 0;
    g_io_channel_read_line(out_ch, &buff, &buff_len, &tpos, &error);
    if ( error )
    {
        g_printerr("Could not start daemon: %s\n", error->message);
        g_error_free(error);
        exit(1);
    }
    g_free(buff);

    g_io_add_watch( out_ch, (GIOCondition)(G_IO_IN | G_IO_HUP), (GIOFunc)cb_out_watch, 0 );
    return pid;
}

//GData *fildes;
std::map<GSocket *, GSocket *> file_map;

/*
 * We connect to a daemon twice per process. One is for the
 * ioctl/lseek operations and the other one is for read/write. We fork
 * a subprocess that will behave as the kernel module.
 */
GSocket *do_open(const char *path, int flags, int firmware, GPid *_pid)
{
    GSocket *fd=0, *cfd=0;
    GPid pid;
    unsigned long handshake;
    GError *error = 0;
    file_map.clear();

    pid = check_for_daemon(firmware);
    if (_pid)
        *_pid = pid;

    // Create the ioctl channel
    int mx_try = 10;
    for (int itry=0; itry<mx_try && cfd==0; itry++)
    {
        cfd = connect_server('c', SOCKET_NAME);
        if (cfd)
            break;

        g_usleep(1000);
    }
    PDEBUG("cfd socket is %lx\n", (unsigned long)cfd);
    if (!cfd)
        return 0;
    else
    {
        int opt_val = 1;
        setsockopt( g_socket_get_fd(cfd),
                    IPPROTO_TCP,
                    TCP_NODELAY, (char *)&opt_val,
                    sizeof(int));
        g_socket_receive(cfd, (gchar *)&handshake, sizeof(unsigned long), 0, &error);
        if (error)
        {
            err_msg(_("Cannot receive handshake word: %s\n"), error->message );
            g_error_free(error);
            error = 0;
        }
    }

    // Create the I/O pipe
    fd = connect_server('m', SOCKET_NAME);
    PDEBUG("fd socket is %lx\n", (unsigned long)fd);
    if (!fd)
        return 0;
    else
    {
        int opt_val = 1;
        setsockopt( g_socket_get_fd(fd),
                    IPPROTO_TCP,
                    TCP_NODELAY, (char *)&opt_val,
                    sizeof(int));

        g_socket_receive(fd, (gchar *)&handshake, sizeof(unsigned long), 0, &error);
        if (error)
        {
            err_msg( _("Cannot receive handshake word: %s\n"), error->message );
            g_error_free(error);
            error = 0;
        }
    }

    file_map[fd] = cfd;

    return fd;
}

/*
 *---------------------------------------------------------------
 * The client routines.
 * These routines are called by the client to communicate with the
 * process that fakes the driver.
 *---------------------------------------------------------------
 */



/*
 * Send the ioctl commands to the "driver"
 */
int do_ioctl(GSocket *fd, int req, ...)
{
    va_list args;
    GError *error=0;
    void *ptr;
    int rc, tmp;
    long lval;
    int dir= _IOC_DIR (req);
    int sz= _IOC_SIZE (req);

    GSocket *cfd = file_map[fd];

    PDEBUG ("sending %x\n", req);

    /*
     * If we want the server to DIE, we do not make any other
     * complicated stuff.
     */
    if (req == DIE)
    {
        PDEBUG ("This is the end\n");
        rc = g_socket_send(cfd, (gchar *)&req, sizeof(int), 0, &error);
        rc = g_socket_receive(cfd, (gchar *)&tmp, sizeof(int), 0, &error);
        return 0;
    }

    // Otherwise, send the request
    rc = g_socket_send(cfd, (gchar *)&req, sizeof(int), 0, &error);
    if ( error )
    {
        err_msg(_("Cannot sent ioctl request: %s"), error->message);
        return -1;
    }
    if (rc < 0)
        return rc;

    // See what else is needed. We can set get or both
    if (dir & (_IOC_READ | _IOC_WRITE))
    {
        va_start(args, req);
        lval = va_arg (args, long);
        ptr = (char *)&lval;

        if (dir & _IOC_WRITE)
        {
            // we have to send a value
            rc = g_socket_send(cfd, (gchar *)ptr, sz, 0, &error);
            if ( error )
            {
                err_msg(_("Cannot sent ioctl request data: %s"), error->message);
                return -1;
            }
            if (rc != sz )
            {
                err_msg(_("Did not send the total ammount: %d (%d)"), sz, rc);
            }
        }
        if (dir & _IOC_READ)
        {
            int rc;
            // we have to read something back
            ptr = (char *) lval;
            rc = g_socket_receive(cfd, (gchar *)ptr, sz, 0, &error);
            if (rc<=0)
            {
                err_msg(_("problems reading back the value"));
            }
        }
        va_end(args);
    }
    // get the return code;
    rc = g_socket_receive(cfd, (gchar *)&tmp, sizeof(int), 0, &error);

    return tmp;
}

/*
 * Reading from the server.
 */
int do_read(GSocket *fd, void *buff, size_t size)
{
    GError *error = 0;
    if (!fd)
        return 0;

    int iok, rc;
    rc = do_ioctl(fd, DO_READ, size);
    iok = g_socket_receive(fd, (gchar *)buff, size, 0, &error);
    if (error)
    {
        err_msg(_("Could not receive data: %s\n"), error->message);
        g_error_free(error);
    }
    rc = do_ioctl(fd, ALIBAVA_ERROR);
    if (rc<0 || iok<=0)
        return -1;

    return iok;
}

/*
 * Write is just a normal write. The server should be able to retrieve
 * the ammount of information with FIONREAD
 */
int do_write(GSocket *fd, void *buff, size_t size)
{
    GError *error = 0;
    if (!fd)
        return 0;


    int rc = g_socket_send(fd, (gchar *)buff, size, 0, &error);
    if ( error )
    {
        rc = - error->code;
        err_msg(_("Error writing to socket: (%d) %s\n"), error->code, error->message);
        g_error_free(error);
    }
    return rc;
}

/*
 * The close operation. Send the DIE message to the daemon and
 * close
 */
int do_close(GSocket *fd)
{
    int rc;
    GError *error = 0;
    GSocket *cfd = file_map[fd];
    rc = do_ioctl(fd, DIE);
    g_usleep(100000);

    g_socket_close(fd, &error);
    g_socket_close(cfd, &error);
    g_object_unref(fd);
    g_object_unref(cfd);

    file_map.erase(fd);
    return 0;
}

struct USBemulator_priv
{
        GSocket *S;
        GIOChannel *chan;
        GPid pid;
        bool use_fifo;
};

USBemulator::USBemulator(int firmware)
    : _firmware(firmware), _priv(0)
{
    _priv = g_new0(USBemulator_priv, 1);
    _priv->S = 0;
    _priv->chan = 0;
    _priv->pid = 0;
    _priv->use_fifo = false;
    fd = 0;
}

USBemulator::~USBemulator()
{
    if (fd>=0)
        close();

    kill_daemon();
    __g_gpid_kill(_priv->pid);
    g_spawn_close_pid (_priv->pid);
    g_free(_priv);
}

int USBemulator::error() const
{

    return _priv->S ? do_ioctl(_priv->S, ERROR_NR) : -1;
}

int USBemulator::open(const std::string &fnam)
{
    _priv->S = do_open(fnam.c_str(), 0, _firmware, &_priv->pid);
    if (!_priv->S)
    {
        err_msg(_("do_open does not return a valid socket"));
        return -1;
    }
    else
    {
        GError *error = 0;
#ifdef G_OS_WIN32
      _priv->chan = g_io_channel_win32_new_socket( g_socket_get_fd(_priv->S));
#else
      fd = g_socket_get_fd(_priv->S);
      _priv->chan = g_io_channel_unix_new( fd );
#endif
      g_io_channel_set_encoding(_priv->chan, 0, &error);
      g_io_channel_set_buffered(_priv->chan, FALSE);
      g_io_channel_ref(_priv->chan);
    }
    return 0;
}

int USBemulator::close()
{
    if ( _priv->S)
    {

        return do_close(_priv->S);
    }
    else
        return 0;
}

int USBemulator::read(void *buf, int size)
{
    return do_read(_priv->S, buf, size);
}

int USBemulator::write(void *buf, int size)
{
    return do_write(_priv->S, buf, size);
}

int USBemulator::poll(int timeout)
{
    GError *err = 0;
    GCancellable *cancellable = g_cancellable_new();
#if GLIB_CHECK_VERSION(2,26,0)
    gint old_timeout = g_socket_get_timeout(_priv->S);
    if (timeout>0)
        g_socket_set_timeout(_priv->S, timeout);
#endif

    gboolean rc = g_socket_condition_wait (_priv->S, G_IO_IN,
                                         cancellable,
                                         &err);

#if GLIB_CHECK_VERSION(2,26,0)
    if (timeout>0)
        g_socket_set_timeout(_priv->S, old_timeout );
#endif

    g_object_unref(cancellable);
    if (rc)
        return 0;
    else {
        return -1;
    }
}

int USBemulator::setup_fifo(Fifo *f)
{
    _priv->use_fifo = true;
    f->attach(_priv->S, _priv->chan);

    return 0;
}
#endif
